@@ -18,3 +18,6 @@ OVS_BRIDGE_INTERFACE = "bridge"
 OVS_PATCH_INTERFACE = "patch"
 OVS_TUNNEL_INTERFACE = "tunnel"
 OVS_UNKNOWN_INTERFACE = "unknown"
+
+PORT_STATUS_UP = "up"
+PORT_STATUS_DOWN = "down"
