@@ -19,54 +19,53 @@ import time
 
 from neutron.agent.common import config
 from neutron.common import config as common_config
-from oslo_config import cfg
 from oslo_log import log
-from oslo_serialization import jsonutils
 from ryu.base import app_manager
 from ryu import cfg as ryu_cfg
-from ryu.ofproto import ofproto_common
 
-from dragonflow._i18n import _LI, _LW, _
-from dragonflow.common import common_params
+from dragonflow._i18n import _LI, _LW
 from dragonflow.common import constants
 from dragonflow.common import utils as df_utils
+from dragonflow import conf as cfg
+from dragonflow.controller import df_db_objects_refresh
 from dragonflow.controller import ryu_base_app
 from dragonflow.controller import topology
 from dragonflow.db import api_nb
+from dragonflow.db import db_consistent
 from dragonflow.db import db_store
-from dragonflow.db.drivers import ovsdb_vswitch_impl
+from dragonflow.ovsdb import vswitch_impl
 
 
-df_ryu_opts = [
-    cfg.IPOpt('of_listen_address', default='127.0.0.1',
-              help=_("Address to listen on for OpenFlow connections.")),
-    cfg.PortOpt('of_listen_port', default=ofproto_common.OFP_TCP_PORT,
-                help=_("Port to listen on for OpenFlow connections."))
-]
-
-config.setup_logging()
 LOG = log.getLogger("dragonflow.controller.df_local_controller")
 
-cfg.CONF.register_opts(common_params.df_opts, 'df')
-cfg.CONF.register_opts(df_ryu_opts, 'df_ryu')
-
 
 class DfLocalController(object):
 
     def __init__(self, chassis_name):
-        self.next_network_id = 0
         self.db_store = db_store.DbStore()
         self.chassis_name = chassis_name
+        self.mgt_ip = cfg.CONF.df.management_ip
         self.ip = cfg.CONF.df.local_ip
-        self.tunnel_type = cfg.CONF.df.tunnel_type
+        if cfg.CONF.df.tunnel_types:
+            # Virtual tunnel port support multiple tunnel types together
+            self.tunnel_types = cfg.CONF.df.tunnel_types
+        else:
+            # NOTE(xiaohhui): This should be removed along with the config
+            # option tunnel_type
+            self.tunnel_types = [cfg.CONF.df.tunnel_type]
         self.sync_finished = False
+        self.port_status_notifier = None
         nb_driver = df_utils.load_driver(
             cfg.CONF.df.nb_db_class,
             df_utils.DF_NB_DB_DRIVER_NAMESPACE)
         self.nb_api = api_nb.NbApi(
             nb_driver,
             use_pubsub=cfg.CONF.df.enable_df_pub_sub)
-        self.vswitch_api = ovsdb_vswitch_impl.OvsdbSwitchApi(self.ip)
+        self.vswitch_api = vswitch_impl.OvsApi(self.mgt_ip)
+        if cfg.CONF.df.enable_port_status_notifier:
+            self.port_status_notifier = df_utils.load_driver(
+                     cfg.CONF.df.port_status_notifier,
+                     df_utils.DF_PORT_STATUS_DRIVER_NAMESPACE)
         kwargs = dict(
             nb_api=self.nb_api,
             vswitch_api=self.vswitch_api,
@@ -75,21 +74,31 @@ class DfLocalController(object):
         app_mgr = app_manager.AppManager.get_instance()
         self.open_flow_app = app_mgr.instantiate(ryu_base_app.RyuDFAdapter,
                                                  **kwargs)
-
         self.topology = None
+        self.db_consistency_manager = None
+        self.enable_db_consistency = cfg.CONF.df.enable_df_db_consistency
         self.enable_selective_topo_dist = \
             cfg.CONF.df.enable_selective_topology_distribution
-        self.remote_chassis_lport_map = {}
         self.integration_bridge = cfg.CONF.df.integration_bridge
 
     def run(self):
         self.nb_api.initialize(db_ip=cfg.CONF.df.remote_db_ip,
                                db_port=cfg.CONF.df.remote_db_port)
         self.vswitch_api.initialize(self.nb_api)
+        if cfg.CONF.df.enable_port_status_notifier:
+            self.port_status_notifier.initialize(mech_driver=None,
+                                             nb_api=self.nb_api,
+                                             pub=self.nb_api.publisher,
+                                             sub=None,
+                                             is_neutron_server=False)
         self.topology = topology.Topology(self,
                                           self.enable_selective_topo_dist)
+        if self.enable_db_consistency:
+            self.db_consistency_manager = \
+                db_consistent.DBConsistencyManager(self)
+            self.nb_api.set_db_consistency_manager(self.db_consistency_manager)
+            self.db_consistency_manager.daemonize()
 
-        self.vswitch_api.sync()
         # both set_controller and del_controller will delete flows.
         # for reliability, here we should check if controller is set for OVS,
         # if yes, don't set controller and don't delete controller.
@@ -105,6 +114,8 @@ class DfLocalController(object):
             self.vswitch_api.set_controller_fail_mode(
                 self.integration_bridge, 'secure')
         self.open_flow_app.start()
+        self.create_tunnels()
+        df_db_objects_refresh.initialize_object_refreshers(self)
         self.db_sync_loop()
 
     def db_sync_loop(self):
@@ -115,8 +126,12 @@ class DfLocalController(object):
                     self.nb_api.support_publish_subscribe()):
                 self.nb_api.register_notification_callback(self)
 
-    def run_sync(self):
-        self.sync_finished = True
+    def run_sync(self, mode=None):
+        if mode == 'full_sync':
+            # For a full sync, df needs to clean the local cache, so that
+            # all resources will be treated as new resource, and thus be
+            # applied to local.
+            self.db_store.clear()
         while True:
             time.sleep(1)
             self.run_db_poll()
@@ -125,80 +140,48 @@ class DfLocalController(object):
 
     def run_db_poll(self):
         try:
-            self.nb_api.sync()
-
             self.register_chassis()
 
-            self.create_tunnels()
-
-            if not self.enable_selective_topo_dist:
-
-                self.read_switches()
-
-                self.read_security_groups()
-
-                self.port_mappings()
-
-                self.read_routers()
-
-                self.read_floatingip()
-
+            topics = self.topology.get_subscribed_topics()
+            df_db_objects_refresh.sync_local_cache_from_nb_db(topics)
             self.sync_finished = True
-
         except Exception as e:
             self.sync_finished = False
             LOG.warning(_LW("run_db_poll - suppressing exception"))
             LOG.exception(e)
 
-    def chassis_created(self, chassis):
-        # Check if tunnel already exists to this chassis
-        t_ports = self.vswitch_api.get_tunnel_ports()
+    def update_chassis(self, chassis):
+        self.db_store.update_chassis(chassis.get_id(), chassis)
         remote_chassis_name = chassis.get_id()
         if self.chassis_name == remote_chassis_name:
             return
-        for t_port in t_ports:
-            if t_port.get_chassis_id() == remote_chassis_name:
-                LOG.info(_LI("remote Chassis Tunnel already installed  = %s") %
-                     chassis.__str__())
-                return
-        # Create tunnel port to this chassis
-        LOG.info(_LI("Adding tunnel to remote chassis = %s") %
-                 chassis.__str__())
-        self.vswitch_api.add_tunnel_port(chassis)
 
-    def chassis_deleted(self, chassis_id):
-        LOG.info(_LI("Deleting tunnel to remote chassis = %s") % chassis_id)
-        tunnel_ports = self.vswitch_api.get_tunnel_ports()
-        for port in tunnel_ports:
-            if port.get_chassis_id() == chassis_id:
-                self.vswitch_api.delete_port(port)
-                return
+        # Notify about remote port update
+        remote_ports = self.db_store.get_ports_by_chassis(remote_chassis_name)
+        for port in remote_ports:
+            self._logical_port_process(port, None)
 
-    def read_switches(self):
-        lswitches_to_remove = self.db_store.get_lswitch_keys()
-        for lswitch in self.nb_api.get_all_logical_switches():
-            self.logical_switch_updated(lswitch)
-            if lswitch.get_id() in lswitches_to_remove:
-                lswitches_to_remove.remove(lswitch.get_id())
+    def delete_chassis(self, chassis_id):
+        LOG.info(_LI("Deleting remote ports in remote chassis %s"), chassis_id)
+        # Chassis is deleted, there is no reason to keep the remote port
+        # in it.
+        remote_ports = self.db_store.get_ports_by_chassis(chassis_id)
+        for port in remote_ports:
+            self.logical_port_deleted(port.get_id())
+        self.db_store.delete_chassis(chassis_id)
 
-        for lswitch_to_remove in lswitches_to_remove:
-            self.logical_switch_deleted(lswitch_to_remove)
-
-    def logical_switch_updated(self, lswitch):
+    def update_lswitch(self, lswitch):
         old_lswitch = self.db_store.get_lswitch(lswitch.get_id())
-        if old_lswitch == lswitch:
+        if not df_utils.is_valid_version(
+                old_lswitch.inner_obj if old_lswitch else None,
+                lswitch.inner_obj):
             return
-        #Make sure we have a local network_id mapped before we dispatch
-        network_id = self.get_network_id(
-            lswitch.get_id(),
-        )
-        lswitch_conf = {'network_id': network_id, 'lswitch':
-            lswitch.__str__()}
-        LOG.info(_LI("Adding/Updating Logical Switch = %s") % lswitch_conf)
+
+        LOG.info(_LI("Adding/Updating Logical Switch = %s"), lswitch)
         self.db_store.set_lswitch(lswitch.get_id(), lswitch)
         self.open_flow_app.notify_update_logical_switch(lswitch)
 
-    def logical_switch_deleted(self, lswitch_id):
+    def delete_lswitch(self, lswitch_id):
         lswitch = self.db_store.get_lswitch(lswitch_id)
         LOG.info(_LI("Removing Logical Switch = %s") % lswitch_id)
         if lswitch is None:
@@ -207,7 +190,22 @@ class DfLocalController(object):
             return
         self.open_flow_app.notify_remove_logical_switch(lswitch)
         self.db_store.del_lswitch(lswitch_id)
-        self.db_store.del_network_id(lswitch_id)
+
+    def _notify_active_ports_updated_when_lport_created(self, lport):
+        active_ports = self.nb_api.get_active_ports(lport.get_topic())
+        for active_port in active_ports:
+            if active_port.get_detected_lport_id() == lport.get_id():
+                self.db_store.update_active_port(active_port.get_id(),
+                                                 active_port)
+                self.open_flow_app.notify_update_active_port(active_port,
+                                                             None)
+
+    def _notify_active_ports_updated_when_lport_removed(self, lport):
+        active_ports = self.db_store.get_active_ports(lport.get_topic())
+        for active_port in active_ports:
+            if active_port.get_detected_lport_id() == lport.get_id():
+                self.open_flow_app.notify_remove_active_port(active_port)
+                self.db_store.delete_active_port(active_port.get_id())
 
     def _is_physical_chassis(self, chassis):
         if not chassis or chassis == constants.DRAGONFLOW_VIRTUAL_PORT:
@@ -215,27 +213,31 @@ class DfLocalController(object):
         return True
 
     def _logical_port_process(self, lport, original_lport=None):
-        chassis = lport.get_chassis()
-        local_network_id = self.get_network_id(
-            lport.get_lswitch_id(),
-        )
         lswitch = self.db_store.get_lswitch(lport.get_lswitch_id())
-        if lswitch is not None:
-            network_type = lswitch.get_network_type()
-            segment_id = lswitch.get_segment_id()
+        if not lswitch:
+            LOG.warning(_LW("Could not find lswitch for lport: %s"),
+                        lport.get_id())
+            return
+        lport.set_external_value('local_network_id',
+                                 lswitch.get_unique_key())
+        network_type = lswitch.get_network_type()
+        segment_id = lswitch.get_segment_id()
+        physical_network = lswitch.get_physical_network()
 
-            lport.set_external_value('network_type', network_type)
-            if segment_id is not None:
-                lport.set_external_value('segmentation_id',
+        lport.set_external_value('network_type', network_type)
+        if segment_id is not None:
+            lport.set_external_value('segmentation_id',
                                      int(segment_id))
-        lport.set_external_value('local_network_id', local_network_id)
+        if physical_network:
+            lport.set_external_value('physical_network', physical_network)
 
+        chassis = lport.get_chassis()
         if chassis == self.chassis_name:
             lport.set_external_value('is_local', True)
+            self.db_store.set_port(lport.get_id(), lport, True)
             ofport = self.vswitch_api.get_port_ofport_by_id(lport.get_id())
             if ofport:
                 lport.set_external_value('ofport', ofport)
-                self.db_store.set_port(lport.get_id(), lport, True)
                 if original_lport is None:
                     LOG.info(_LI("Adding new local logical port = %s") %
                              str(lport))
@@ -250,12 +252,27 @@ class DfLocalController(object):
             else:
                 LOG.info(_LI("Local logical port %s was not created yet") %
                          str(lport))
+                return
         else:
             lport.set_external_value('is_local', False)
-            ofport = self.vswitch_api.get_chassis_ofport(chassis)
+            self.db_store.set_port(lport.get_id(), lport, False)
+            if lport.get_remote_vtep():
+                # Remote port that exists in other network pod.
+                lport.set_external_value('peer_vtep_address',
+                                         lport.get_chassis())
+            else:
+                # Remote port that exists in current network pod.
+                remote_chassis = self.db_store.get_chassis(lport.get_chassis())
+                if not remote_chassis:
+                    # chassis has not been online yet.
+                    return
+                lport.set_external_value('peer_vtep_address',
+                                         remote_chassis.get_ip())
+
+            ofport = self.vswitch_api.get_vtp_ofport(
+                lport.get_external_value('network_type'))
             if ofport:
                 lport.set_external_value('ofport', ofport)
-                self.db_store.set_port(lport.get_id(), lport, False)
                 if original_lport is None:
                     LOG.info(_LI("Adding new remote logical port = %s") %
                              str(lport))
@@ -268,65 +285,32 @@ class DfLocalController(object):
                     self.open_flow_app.notify_update_remote_port(
                         lport, original_lport)
             else:
-                # TODO(gampel) add handling for this use case
-                # remote port but no tunnel to remote Host
-                # if this should never happen raise an exception
-                LOG.warning(_LW("No tunnel for remote logical port %s") %
-                            str(lport))
+                # The tunnel port online event will update the remote logical
+                # port. Log this warning first.
+                LOG.warning(_LW("No tunnel for remote logical port %s"),
+                            lport)
+                return
 
-    def _add_remote_port_on_chassis(self, lport):
-        chassis = lport.get_chassis()
-        chassis_lports = self.remote_chassis_lport_map.get(chassis)
-        if not chassis_lports:
-            chassis_value = {'id': chassis, 'ip': chassis,
-                             'tunnel_type': self.tunnel_type}
-            chassis_inst = api_nb.Chassis(jsonutils.dumps(chassis_value))
-            self.chassis_created(chassis_inst)
-            self.remote_chassis_lport_map[chassis] = {lport.get_id()}
-        else:
-            chassis_lports.add(lport.get_id())
+        if original_lport is None:
+            self._notify_active_ports_updated_when_lport_created(lport)
 
-    def _delete_remote_port_from_chassis(self, lport):
-        chassis = lport.get_chassis()
-        chassis_lports = self.remote_chassis_lport_map.get(chassis)
-        if chassis_lports is None:
-            return
-        chassis_lports.remove(lport.get_id())
-        if len(chassis_lports) == 0:
-            self.chassis_deleted(chassis)
-            del self.remote_chassis_lport_map[chassis]
-
-    def logical_port_created(self, lport):
-        chassis = lport.get_chassis()
-        if not self._is_physical_chassis(chassis):
-            LOG.debug(("Port %s has not been bound or it is a vPort") %
-                      lport.get_id())
-            return
-        if lport.get_remote_vtep():
-            self._add_remote_port_on_chassis(lport)
-        self._logical_port_process(lport)
-
-    def logical_port_updated(self, lport):
+    def update_lport(self, lport):
         chassis = lport.get_chassis()
         if not self._is_physical_chassis(chassis):
             LOG.debug(("Port %s has not been bound or it is a vPort") %
                       lport.get_id())
             return
         original_lport = self.db_store.get_port(lport.get_id())
-        if not original_lport:
-            if lport.get_remote_vtep():
-                self._add_remote_port_on_chassis(lport)
-        else:
-            original_chassis = original_lport.get_chassis()
-            if original_chassis != chassis:
-                if original_lport.get_remote_vtep():
-                    self._delete_remote_port_from_chassis(original_lport)
+        if original_lport and not original_lport.get_external_value("ofport"):
+            original_lport = None
 
-                if lport.get_remote_vtep():
-                    self._add_remote_port_on_chassis(lport)
+        if not df_utils.is_valid_version(
+                original_lport.inner_obj if original_lport else None,
+                lport.inner_obj):
+            return
         self._logical_port_process(lport, original_lport)
 
-    def logical_port_deleted(self, lport_id):
+    def delete_lport(self, lport_id):
         lport = self.db_store.get_port(lport_id)
         if lport is None:
             return
@@ -343,177 +327,107 @@ class DfLocalController(object):
                 self.open_flow_app.notify_remove_remote_port(lport)
             self.db_store.delete_port(lport.get_id(), False)
 
-        if lport.get_remote_vtep():
-            self._delete_remote_port_from_chassis(lport)
+        self._notify_active_ports_updated_when_lport_removed(lport)
 
     def bridge_port_updated(self, lport):
         self.open_flow_app.notify_update_bridge_port(lport)
 
-    def router_updated(self, lrouter):
+    def update_lrouter(self, lrouter):
         old_lrouter = self.db_store.get_router(lrouter.get_id())
-        if old_lrouter is None:
-            LOG.info(_LI("Logical Router created = %s") %
-                     lrouter.__str__())
-            self._add_new_lrouter(lrouter)
+        if not df_utils.is_valid_version(
+                old_lrouter.inner_obj if old_lrouter else None,
+                lrouter.inner_obj):
             return
-        self._update_router_interfaces(old_lrouter, lrouter)
-        self._update_router_attributes(old_lrouter, lrouter)
+        self.open_flow_app.notify_update_router(lrouter, old_lrouter)
         self.db_store.update_router(lrouter.get_id(), lrouter)
 
-    def router_deleted(self, lrouter_id):
-        old_lrouter = self.db_store.get_router(lrouter_id)
-        if old_lrouter is None:
+    def delete_lrouter(self, lrouter_id):
+        router = self.db_store.get_router(lrouter_id)
+        if router is None:
+            LOG.warning(_LW("Try to delete a nonexistent router(%s)"),
+                        lrouter_id)
             return
-        old_router_ports = old_lrouter.get_ports()
-        for old_port in old_router_ports:
-            self._delete_router_port(old_port)
+        LOG.info(_LI("Removing router = %s"), lrouter_id)
+        self.open_flow_app.notify_delete_router(router)
         self.db_store.delete_router(lrouter_id)
 
-    def security_group_updated(self, secgroup):
+    def update_secgroup(self, secgroup):
         old_secgroup = self.db_store.get_security_group(secgroup.get_id())
         if old_secgroup is None:
             LOG.info(_LI("Security Group created = %s") %
                      secgroup)
             self._add_new_security_group(secgroup)
             return
+        if not df_utils.is_valid_version(
+                old_secgroup.inner_obj if old_secgroup else None,
+                secgroup.inner_obj):
+            return
         self._update_security_group_rules(old_secgroup, secgroup)
         self.db_store.update_security_group(secgroup.get_id(), secgroup)
 
-    def security_group_deleted(self, secgroup_id):
+    def delete_secgroup(self, secgroup_id):
         old_secgroup = self.db_store.get_security_group(secgroup_id)
         if old_secgroup is None:
             return
         self._delete_old_security_group(old_secgroup)
 
-    def register_chassis(self):
-        chassis = self.nb_api.get_chassis(self.chassis_name)
-        # TODO(gsagie) Support tunnel type change here ?
+    def update_qospolicy(self, qos):
+        original_qos = self.db_store.get_qos_policy(qos.get_id())
+        if not df_utils.is_valid_version(
+                original_qos.inner_obj if original_qos else None,
+                qos.inner_obj):
+            return
 
+        self.db_store.set_qos_policy(qos.get_id(), qos)
+        if not original_qos:
+            return
+
+        self.open_flow_app.notify_update_qos_policy(qos)
+
+    def delete_qospolicy(self, qos_id):
+        qos = self.db_store.get_qos_policy(qos_id)
+        if not qos:
+            return
+
+        self.open_flow_app.notify_delete_qos_policy(qos)
+        self.db_store.delete_qos_policy(qos_id)
+
+    def register_chassis(self):
+        # Get all chassis from nb db to db store.
+        for c in self.nb_api.get_all_chassis():
+            self.db_store.update_chassis(c.get_id(), c)
+
+        chassis = self.db_store.get_chassis(self.chassis_name)
         if chassis is None:
             self.nb_api.add_chassis(self.chassis_name,
                                     self.ip,
-                                    self.tunnel_type)
+                                    self.tunnel_types)
+        else:
+            kwargs = {}
+            old_tunnel_types = chassis.get_tunnel_types()
+            if (not isinstance(old_tunnel_types, list) or
+                    set(self.tunnel_types) != set(old_tunnel_types)):
+                # There are 2 cases that needs update tunnel type in
+                # chassis. 1) User changes tunnel types in conf file
+                # 2) An old controller support only one type tunnel switch
+                # to support virtual tunnel port.
+                kwargs['tunnel_types'] = self.tunnel_types
+            if self.ip != chassis.get_ip():
+                kwargs['ip'] = self.ip
+
+            if kwargs:
+                self.nb_api.update_chassis(self.chassis_name, **kwargs)
 
     def create_tunnels(self):
-        tunnel_ports = {}
-        t_ports = self.vswitch_api.get_tunnel_ports()
-        for t_port in t_ports:
-            tunnel_ports[t_port.get_chassis_id()] = t_port
+        tunnel_ports = self.vswitch_api.get_virtual_tunnel_ports()
+        for tunnel_port in tunnel_ports:
+            if tunnel_port.get_tunnel_type() not in self.tunnel_types:
+                self.vswitch_api.delete_port(tunnel_port)
 
-        for chassis in self.nb_api.get_all_chassis():
-            if chassis.get_id() in tunnel_ports:
-                del tunnel_ports[chassis.get_id()]
-            elif chassis.get_id() == self.chassis_name:
-                pass
-            else:
-                self.chassis_created(chassis)
-
-        # Iterate all tunnel ports that needs to be deleted
-        for port in tunnel_ports.values():
-            self.vswitch_api.delete_port(port)
-
-    def port_mappings(self):
-        ports_to_remove = self.db_store.get_port_keys()
-        for lport in self.nb_api.get_all_logical_ports():
-            self.logical_port_updated(lport)
-            if lport.get_id() in ports_to_remove:
-                ports_to_remove.remove(lport.get_id())
-
-        for port_to_remove in ports_to_remove:
-            self.logical_port_deleted(port_to_remove)
-
-    def get_network_id(self, logical_dp_id):
-        network_id = self.db_store.get_network_id(logical_dp_id)
-        if network_id is not None:
-            return network_id
-        else:
-            self.next_network_id += 1
-            # TODO(gsagie) verify self.next_network_id didnt wrap
-            self.db_store.set_network_id(
-                logical_dp_id,
-                self.next_network_id,
-            )
-            return self.next_network_id
-
-    def read_routers(self):
-        routers_to_remove = self.db_store.get_router_keys()
-        for lrouter in self.nb_api.get_routers():
-            self.router_updated(lrouter)
-            if lrouter.get_id() in routers_to_remove:
-                routers_to_remove.remove(lrouter.get_id())
-
-        for router_to_remove in routers_to_remove:
-            self.router_deleted(router_to_remove)
-
-    def _update_router_attributes(self, old_router, new_router):
-        old_routes = old_router.get_routes()
-        new_routes = new_router.get_routes()
-        for new_route in new_routes:
-            if new_route not in old_routes:
-                self._add_router_route(new_router, new_route)
-            else:
-                old_routes.remove(new_route)
-        for old_route in old_routes:
-            self._delete_router_route(new_router, old_route)
-
-    def _add_router_route(self, router, route):
-        LOG.info(_LI("Adding new logical router route = %s"), route)
-        self.open_flow_app.notify_add_router_route(
-            router, route)
-
-    def _delete_router_route(self, router, route):
-        LOG.info(_LI("Removing logical router route = %s"), route)
-        self.open_flow_app.notify_remove_router_route(
-            router, route)
-
-    def _update_router_interfaces(self, old_router, new_router):
-        new_router_ports = new_router.get_ports()
-        old_router_ports = old_router.get_ports()
-        for new_port in new_router_ports:
-            if new_port not in old_router_ports:
-                self._add_new_router_port(new_router, new_port)
-            else:
-                old_router_ports.remove(new_port)
-
-        for old_port in old_router_ports:
-            self._delete_router_port(old_port)
-
-    def _add_new_router_port(self, router, router_port):
-        LOG.info(_LI("Adding new logical router interface = %s") %
-                 router_port.__str__())
-        local_network_id = self.db_store.get_network_id(
-            router_port.get_lswitch_id()
-        )
-        self.open_flow_app.notify_add_router_port(
-                router, router_port, local_network_id)
-
-    def _delete_router_port(self, router_port):
-        LOG.info(_LI("Removing logical router interface = %s") %
-                 router_port.__str__())
-        local_network_id = self.db_store.get_network_id(
-            router_port.get_lswitch_id()
-        )
-        self.open_flow_app.notify_remove_router_port(
-                router_port, local_network_id)
-
-    def _add_new_lrouter(self, lrouter):
-        for new_port in lrouter.get_ports():
-            self._add_new_router_port(lrouter, new_port)
-        routes = lrouter.get_routes()
-        for route in routes:
-            self._add_router_route(lrouter, route)
-        self.db_store.update_router(lrouter.get_id(), lrouter)
-
-    def read_security_groups(self):
-        secgroups_to_remove = self.db_store.get_security_group_keys()
-
-        for secgroup in self.nb_api.get_security_groups():
-            self.security_group_updated(secgroup)
-            if secgroup.get_id() in secgroups_to_remove:
-                secgroups_to_remove.remove(secgroup.get_id())
-
-        for secgroup_to_remove in secgroups_to_remove:
-            self.security_group_deleted(secgroup_to_remove)
+        for t in self.tunnel_types:
+            # The customized ovs idl will ingore the command if the port
+            # already exists.
+            self.vswitch_api.add_virtual_tunnel_port(t)
 
     def _update_security_group_rules(self, old_secgroup, new_secgroup):
         new_secgroup_rules = new_secgroup.get_rules()
@@ -549,17 +463,7 @@ class DfLocalController(object):
         self.open_flow_app.notify_remove_security_group_rule(
                  secgroup, secgroup_rule)
 
-    def read_floatingip(self):
-        fips_to_remove = self.db_store.get_floatingip_keys()
-        for floatingip in self.nb_api.get_floatingips():
-            self.floatingip_updated(floatingip)
-            if floatingip.get_id() in fips_to_remove:
-                fips_to_remove.remove(floatingip.get_id())
-
-        for fip_to_remove in fips_to_remove:
-            self.floatingip_deleted(fip_to_remove)
-
-    def floatingip_updated(self, floatingip):
+    def update_floatingip(self, floatingip):
         # check whether this floatingip is associated with a lport or not
         if floatingip.get_lport_id():
             if self.db_store.get_local_port(floatingip.get_lport_id()) is None:
@@ -572,22 +476,27 @@ class DfLocalController(object):
                 return
             self._associate_floatingip(floatingip)
             return
+        if not df_utils.is_valid_version(
+                old_floatingip.inner_obj if old_floatingip else None,
+                floatingip.inner_obj):
+            return
         self._update_floatingip(old_floatingip, floatingip)
 
-    def floatingip_deleted(self, floatingip_id):
+    def delete_floatingip(self, floatingip_id):
         floatingip = self.db_store.get_floatingip(floatingip_id)
         if not floatingip:
             return
         self.open_flow_app.notify_delete_floatingip(floatingip)
         LOG.info(_LI("Floatingip is deleted. Floatingip = %s") %
                  str(floatingip))
+        self.db_store.delete_floatingip(floatingip_id)
 
-    def publisher_updated(self, publisher):
+    def update_publisher(self, publisher):
         self.db_store.update_publisher(publisher.get_id(), publisher)
         LOG.info(_LI('Registering to new publisher: %s'), str(publisher))
         self.nb_api.subscriber.register_listen_address(publisher.get_uri())
 
-    def publisher_deleted(self, uuid):
+    def delete_publisher(self, uuid):
         publisher = self.db_store.get_publisher(uuid)
         if publisher:
             LOG.info(_LI('Deleting publisher: %s'), str(publisher))
@@ -628,9 +537,41 @@ class DfLocalController(object):
     def ovs_sync_started(self):
         self.open_flow_app.notify_ovs_sync_started()
 
+    def update_activeport(self, active_port):
+        old_active_port = self.db_store.get_active_port(active_port.get_id())
+        lport_id = active_port.get_detected_lport_id()
+        lport = self.db_store.get_local_port(lport_id,
+                                             active_port.get_topic())
+        LOG.info(_LI("Active port updated. Active port = %(new)s, "
+                     "old active port = %(old)s"),
+                 {'new': active_port, 'old': old_active_port})
+        if lport is not None:
+            self.db_store.update_active_port(active_port.get_id(),
+                                             active_port)
+            self.open_flow_app.notify_update_active_port(active_port,
+                                                         old_active_port)
+        else:
+            LOG.info(_LI("The logical port is not ready for the "
+                         "active node: %s"), active_port)
+
+    def delete_activeport(self, active_port_key):
+        active_port = self.db_store.get_active_port(active_port_key)
+        if active_port is not None:
+            self.db_store.delete_active_port(active_port_key)
+            LOG.info(_LI("Active node was removed. Active node = %s"),
+                     active_port)
+            lport_id = active_port.get_detected_lport_id()
+            lport = self.db_store.get_local_port(lport_id,
+                                                 active_port.get_topic())
+            if lport is not None:
+                self.open_flow_app.notify_remove_active_port(active_port)
+
     def get_nb_api(self):
         return self.nb_api
 
+    def get_portstatus_notifier(self):
+        return self.port_status_notifier
+
     def get_db_store(self):
         return self.db_store
 
@@ -653,6 +594,7 @@ def init_ryu_config():
 def main():
     chassis_name = socket.gethostname()
     common_config.init(sys.argv[1:])
+    config.setup_logging()
     init_ryu_config()
     controller = DfLocalController(chassis_name)
     controller.run()
