@@ -10,10 +10,13 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from oslo_config import cfg
 from oslo_log import log
 
 from dragonflow._i18n import _LI, _LE, _LW
-from dragonflow.db import api_nb
+from dragonflow.common import constants
+from dragonflow.controller import df_db_objects_refresh
+from dragonflow.db import models as db_models
 
 LOG = log.getLogger(__name__)
 
@@ -21,10 +24,10 @@ LOG = log.getLogger(__name__)
 class Topology(object):
 
     def __init__(self, controller, enable_selective_topology_distribution):
-        self.ovs_port_type = (api_nb.OvsPort.TYPE_VM,
-                              api_nb.OvsPort.TYPE_TUNNEL,
-                              api_nb.OvsPort.TYPE_PATCH,
-                              api_nb.OvsPort.TYPE_BRIDGE)
+        self.ovs_port_type = (db_models.OvsPort.TYPE_VM,
+                              db_models.OvsPort.TYPE_TUNNEL,
+                              db_models.OvsPort.TYPE_PATCH,
+                              db_models.OvsPort.TYPE_BRIDGE)
 
         # Stores topics(tenants) subscribed by lports in the current local
         # controller. I,e, {tenant1:{lport1, lport2}, tenant2:{lport3}}
@@ -36,6 +39,7 @@ class Topology(object):
 
         self.controller = controller
         self.nb_api = controller.get_nb_api()
+        self.port_status_reporter = controller.get_portstatus_notifier()
         self.db_store = controller.get_db_store()
         self.openflow_app = controller.get_openflow_app()
         self.chassis_name = controller.get_chassis_name()
@@ -124,14 +128,52 @@ class Topology(object):
         if (ofport is None) or (ofport < 0) or (port_type is None):
             return False
 
-        if (port_type == api_nb.OvsPort.TYPE_VM) \
-                and (ovs_port.get_iface_id() is None):
+        if (port_type == db_models.OvsPort.TYPE_VM and
+                ovs_port.get_iface_id() is None):
             return False
 
         return True
 
+    def _tunnel_port_added(self, ovs_port):
+        self._tunnel_port_updated(ovs_port)
+
+    def _process_ovs_tunnel_port(self, ovs_port, action):
+        tunnel_type = ovs_port.get_tunnel_type()
+        if not tunnel_type:
+            return
+
+        lswitch_ids = self.db_store.get_lswitch_keys_by_network_type(
+            tunnel_type)
+        for lswitch_id in lswitch_ids:
+            lports = self.db_store.get_ports_by_network_id(lswitch_id)
+            for lport in lports:
+                if lport.get_external_value('is_local'):
+                    continue
+
+                # Update of virtual tunnel port should update remote port in
+                # the lswitch of same type.
+                try:
+                    if action == "set":
+                        self.controller.update_lport(lport)
+                    else:
+                        self.controller.delete_lport(lport.get_id())
+                except Exception:
+                    LOG.exception(_LE("Failed to process logical port"
+                                      "when %(action)s tunnel %(lport)s"),
+                                  {'action': action, 'lport': lport})
+
+    def _tunnel_port_updated(self, ovs_port):
+        self._process_ovs_tunnel_port(ovs_port, "set")
+
+    def _tunnel_port_deleted(self, ovs_port):
+        self._process_ovs_tunnel_port(ovs_port, "delete")
+
     def _vm_port_added(self, ovs_port):
         self._vm_port_updated(ovs_port)
+        # publish vm port up event
+        if cfg.CONF.df.enable_port_status_notifier:
+            self.port_status_reporter.notify_port_status(
+                ovs_port, constants.PORT_STATUS_UP)
 
     def _vm_port_updated(self, ovs_port):
         lport_id = ovs_port.get_iface_id()
@@ -141,19 +183,25 @@ class Topology(object):
                         % str(ovs_port))
             return
         topic = lport.get_topic()
+        if not topic:
+            return
         self._add_to_topic_subscribed(topic, lport_id)
 
-        # update lport, notify apps
         ovs_port_id = ovs_port.get_id()
         self.ovs_to_lport_mapping[ovs_port_id] = {'lport_id': lport_id,
                                                   'topic': topic}
-        LOG.info(_LI("A local logical port(%s) is online") % str(lport))
 
-        try:
-            self.controller.logical_port_updated(lport)
-        except Exception:
-            LOG.exception(_LE('Failed to process logical port online '
-                              'event: %s') % str(lport))
+        cached_lport = self.db_store.get_port(lport_id)
+        if not cached_lport or not cached_lport.get_external_value("ofport"):
+            # If the logical port is not in db store or its ofport is not
+            # valid. It has not been applied to dragonflow apps. We need to
+            # update it in dragonflow controller.
+            LOG.info(_LI("A local logical port(%s) is online"), lport)
+            try:
+                self.controller.update_lport(lport)
+            except Exception:
+                LOG.exception(_LE('Failed to process logical port online '
+                                  'event: %s'), lport)
 
     def _bridge_port_added(self, ovs_port):
         self._bridge_port_updated(ovs_port)
@@ -182,35 +230,34 @@ class Topology(object):
 
         LOG.info(_LI("The logical port(%s) is offline") % str(lport))
         try:
-            self.controller.logical_port_deleted(lport_id)
+            self.controller.delete_lport(lport_id)
         except Exception:
             LOG.exception(_LE(
                 'Failed to process logical port offline event %s') % lport_id)
         finally:
-            # TODO(duankebo) publish vm port offline later
-            # currently we will not publish vm port offline event.
-            # lport = self.nb_api.get_logical_port(lport_id)
-            # if lport.get_chassis() == self.chassis_name:
-            #    self.nb_api.update_lport(lport.get_id(), chassis=None,
-            #                             status='DOWN')
+            # publish vm port down event.
+            if cfg.CONF.df.enable_port_status_notifier:
+                self.port_status_reporter.notify_port_status(
+                    ovs_port, constants.PORT_STATUS_DOWN)
+
             del self.ovs_to_lport_mapping[ovs_port_id]
             self._del_from_topic_subscribed(topic, lport_id)
 
     def _add_to_topic_subscribed(self, topic, lport_id):
-        if not self.enable_selective_topo_dist:
+        if not self.enable_selective_topo_dist or not topic:
             return
 
         if topic not in self.topic_subscribed:
             LOG.info(_LI("Subscribe topic: %(topic)s by lport: %(id)s") %
                      {"topic": topic, "id": lport_id})
             self.nb_api.subscriber.register_topic(topic)
-            self._pull_tenant_topology_from_db(topic, lport_id)
+            self._pull_tenant_topology_from_db(topic)
             self.topic_subscribed[topic] = set([lport_id])
         else:
             self.topic_subscribed[topic].add(lport_id)
 
     def _del_from_topic_subscribed(self, topic, lport_id):
-        if not self.enable_selective_topo_dist:
+        if not self.enable_selective_topo_dist or not topic:
             return
         port_ids = self.topic_subscribed[topic]
         port_ids.remove(lport_id)
@@ -221,54 +268,20 @@ class Topology(object):
             self.nb_api.subscriber.unregister_topic(topic)
             self._clear_tenant_topology(topic)
 
-    def _pull_tenant_topology_from_db(self, tenant_id, lport_id):
-        switches = self.nb_api.get_all_logical_switches(tenant_id)
-        for switch in switches:
-            self.controller.logical_switch_updated(switch)
+    def get_subscribed_topics(self):
+        if not self.enable_selective_topo_dist:
+            # Just return None when enable_selective_topo_dist is False
+            return
 
-        sg_groups = self.nb_api.get_security_groups(tenant_id)
-        for sg_group in sg_groups:
-            self.controller.security_group_updated(sg_group)
+        # Return the actual topics that are subscribed. It could be empty
+        # set, which represents no topic is subscribed now.
+        return set(self.topic_subscribed)
 
-        ports = self.nb_api.get_all_logical_ports(tenant_id)
-        for port in ports:
-            if port.get_id() == lport_id:
-                continue
-            self.controller.logical_port_updated(port)
-
-        routers = self.nb_api.get_routers(tenant_id)
-        for router in routers:
-            self.controller.router_updated(router)
-
-        floating_ips = self.nb_api.get_floatingips(tenant_id)
-        for floating_ip in floating_ips:
-            self.controller.floatingip_updated(floating_ip)
+    def _pull_tenant_topology_from_db(self, tenant_id):
+        df_db_objects_refresh.sync_local_cache_from_nb_db({tenant_id})
 
     def _clear_tenant_topology(self, tenant_id):
-        switches = self.db_store.get_lswitchs()
-        for switch in switches:
-            if tenant_id == switch.get_topic():
-                self.controller.logical_switch_deleted(switch.get_id())
-
-        ports = self.db_store.get_ports()
-        for port in ports:
-            if tenant_id == port.get_topic():
-                self.controller.logical_port_deleted(port.get_id())
-
-        floating_ips = self.db_store.get_floatingips()
-        for floating_ip in floating_ips:
-            if tenant_id == floating_ip.get_topic():
-                self.controller.floatingip_deleted(floating_ip.get_id())
-
-        routers = self.db_store.get_routers()
-        for router in routers:
-            if tenant_id == router.get_topic():
-                self.controller.router_deleted(router.get_id())
-
-        sg_groups = self.db_store.get_security_groups()
-        for sg_group in sg_groups:
-            if tenant_id == sg_group.get_topic():
-                self.controller.security_group_deleted(sg_group.get_id())
+        df_db_objects_refresh.clear_local_cache({tenant_id})
 
     def _get_lport(self, port_id, topic=None):
         lport = self.db_store.get_port(port_id)
@@ -276,3 +289,40 @@ class Topology(object):
             lport = self.nb_api.get_logical_port(port_id, topic)
 
         return lport
+
+    def check_topology_info(self):
+        """
+        In order to prevent the situation that the connection between
+        df controller and df db break down, we should recheck the local
+        ovs ports to make sure all the topics of these ovs ports could
+        be subscribed and all the vms could work well.
+        """
+        new_ovs_to_lport_mapping = {}
+        add_ovs_to_lport_mapping = {}
+        delete_ovs_to_lport_mapping = self.ovs_to_lport_mapping
+        for key, ovs_port in self.ovs_ports.items():
+            if ovs_port.get_type() == db_models.OvsPort.TYPE_VM:
+                lport_id = ovs_port.get_iface_id()
+                lport = self._get_lport(lport_id)
+                if lport is None:
+                    LOG.warning(_LW("No logical port found for ovs port: %s"),
+                                ovs_port)
+                    continue
+                topic = lport.get_topic()
+                if not topic:
+                    continue
+                new_ovs_to_lport_mapping[key] = {
+                    'lport_id': lport_id, 'topic': topic}
+                if not delete_ovs_to_lport_mapping.pop(key, None):
+                    add_ovs_to_lport_mapping[key] = {
+                        'lport_id': lport_id, 'topic': topic}
+        self.ovs_to_lport_mapping = new_ovs_to_lport_mapping
+        for value in add_ovs_to_lport_mapping.values():
+            lport_id = value['lport_id']
+            topic = value['topic']
+            self._add_to_topic_subscribed(topic, lport_id)
+
+        for value in delete_ovs_to_lport_mapping.values():
+            lport_id = value['lport_id']
+            topic = value['topic']
+            self._del_from_topic_subscribed(topic, lport_id)
