@@ -15,7 +15,6 @@ import re
 from oslo_log import log
 from redis import client as redis_client
 from redis import exceptions
-import six
 
 from dragonflow._i18n import _LE, _LW
 from dragonflow.db import db_api
@@ -33,6 +32,7 @@ class RedisDbDriver(db_api.DbApi):
         self.clients = {}
         self.remote_server_lists = []
         self.redis_mgt = None
+        self.is_neutron_server = False
 
     def initialize(self, db_ip, db_port, **args):
         # get remote ip port list
@@ -57,8 +57,8 @@ class RedisDbDriver(db_api.DbApi):
         pass
 
     def delete_table(self, table):
-        local_key = self.uuid_to_key(table, '*', '*')
-        for host, client in six.iteritems(self.clients):
+        local_key = self._uuid_to_key(table, '*', '*')
+        for host, client in self.clients.items():
             local_keys = client.keys(local_key)
             if len(local_keys) > 0:
                 for tmp_key in local_keys:
@@ -66,7 +66,7 @@ class RedisDbDriver(db_api.DbApi):
                         self._execute_cmd("DEL", tmp_key)
                     except Exception:
                         LOG.exception(_LE("exception when delete_table: "
-                              "%(key)s "), {'key': local_key})
+                                          "%(key)s "), {'key': local_key})
 
     def _handle_db_conn_error(self, ip_port, local_key=None):
         self.redis_mgt.remove_node_from_master_list(ip_port)
@@ -76,6 +76,13 @@ class RedisDbDriver(db_api.DbApi):
             LOG.exception(_LE("update server list, key: %(key)s")
                           % {'key': local_key})
 
+    def _sync_master_list(self):
+        if self.is_neutron_server:
+            result = self.redis_mgt.redis_get_master_list_from_syncstring(
+                redis_mgt.RedisMgt.global_sharedlist.raw)
+            if result:
+                self._update_server_list()
+
     def _gen_args(self, local_key, value):
         args = []
         args.append(local_key)
@@ -90,6 +97,12 @@ class RedisDbDriver(db_api.DbApi):
 
         return False
 
+    def _update_client(self, local_key):
+        self._sync_master_list()
+        ip_port = self.redis_mgt.get_ip_by_key(local_key)
+        client = self._get_client(local_key, ip_port)
+        return client
+
     def _execute_cmd(self, oper, local_key, value=None):
         if not self._is_oper_valid(oper):
             LOG.warning(_LW("invalid oper: %(oper)s")
@@ -105,6 +118,7 @@ class RedisDbDriver(db_api.DbApi):
 
         ttl = self.RequestRetryTimes
         asking = False
+        alreadysync = False
         while ttl > 0:
             ttl -= 1
             try:
@@ -114,11 +128,19 @@ class RedisDbDriver(db_api.DbApi):
 
                 return client.execute_command(oper, *arg)
             except exceptions.ConnectionError as e:
+                if not alreadysync:
+                    client = self._update_client(local_key)
+                    alreadysync = True
+                    continue
                 self._handle_db_conn_error(ip_port, local_key)
                 LOG.exception(_LE("connection error while sending "
                                   "request to db: %(e)s") % {'e': e})
                 raise e
             except exceptions.ResponseError as e:
+                if not alreadysync:
+                    client = self._update_client(local_key)
+                    alreadysync = True
+                    continue
                 resp = str(e).split(' ')
                 if 'ASK' in resp[0]:
                     # one-time flag to force a node to serve a query about an
@@ -141,16 +163,21 @@ class RedisDbDriver(db_api.DbApi):
                                   % {'e': e})
                     raise e
             except Exception as e:
+                if not alreadysync:
+                    client = self._update_client(local_key)
+                    alreadysync = True
+                    continue
                 self._handle_db_conn_error(ip_port, local_key)
                 LOG.exception(_LE("exception while sending request to "
                                   "db: %(e)s") % {'e': e})
                 raise e
 
     def get_key(self, table, key, topic=None):
-        if topic is None:
-            local_key = self.uuid_to_key(table, key, '*')
+        if not topic:
+            local_key = self._uuid_to_key(table, key, '*')
+            self._sync_master_list()
             try:
-                for host, client in six.iteritems(self.clients):
+                for host, client in self.clients.items():
                     local_keys = client.keys(local_key)
                     if len(local_keys) == 1:
                         return self._execute_cmd("GET", local_keys[0])
@@ -159,7 +186,7 @@ class RedisDbDriver(db_api.DbApi):
                               % {'key': local_key})
 
         else:
-            local_key = self.uuid_to_key(table, key, topic)
+            local_key = self._uuid_to_key(table, key, topic)
             try:
                 # return nil if not found
                 return self._execute_cmd("GET", local_key)
@@ -168,7 +195,7 @@ class RedisDbDriver(db_api.DbApi):
                               % {'key': local_key})
 
     def set_key(self, table, key, value, topic=None):
-        local_key = self.uuid_to_key(table, key, topic)
+        local_key = self._uuid_to_key(table, key, topic)
 
         try:
             res = self._execute_cmd("SET", local_key, value)
@@ -185,7 +212,7 @@ class RedisDbDriver(db_api.DbApi):
 
     def delete_key(self, table, key, topic=None):
         local_topic = topic
-        local_key = self.uuid_to_key(table, key, local_topic)
+        local_key = self._uuid_to_key(table, key, local_topic)
 
         try:
             res = self._execute_cmd("DEL", local_key)
@@ -200,10 +227,11 @@ class RedisDbDriver(db_api.DbApi):
     def get_all_entries(self, table, topic=None):
         res = []
         ip_port = None
-        if topic is None:
-            local_key = self.uuid_to_key(table, '*', '*')
+        self._sync_master_list()
+        if not topic:
+            local_key = self._uuid_to_key(table, '*', '*')
             try:
-                for host, client in six.iteritems(self.clients):
+                for host, client in self.clients.items():
                     local_keys = client.keys(local_key)
                     if len(local_keys) > 0:
                         for tmp_key in local_keys:
@@ -215,7 +243,7 @@ class RedisDbDriver(db_api.DbApi):
                               % {'key': local_key})
 
         else:
-            local_key = self.uuid_to_key(table, '*', topic)
+            local_key = self._uuid_to_key(table, '*', topic)
             try:
                 ip_port = self.redis_mgt.get_ip_by_key(local_key)
                 client = self._get_client(local_key, ip_port)
@@ -234,10 +262,11 @@ class RedisDbDriver(db_api.DbApi):
     def get_all_keys(self, table, topic=None):
         res = []
         ip_port = None
-        if topic is None:
-            local_key = self.uuid_to_key(table, '*', '*')
+        self._sync_master_list()
+        if not topic:
+            local_key = self._uuid_to_key(table, '*', '*')
             try:
-                for host, client in six.iteritems(self.clients):
+                for host, client in self.clients.items():
                     ip_port = host
                     res.extend(client.keys(local_key))
                 return [self._strip_table_name_from_key(key) for key in res]
@@ -248,7 +277,7 @@ class RedisDbDriver(db_api.DbApi):
                               % {'key': local_key, 'e': e})
 
         else:
-            local_key = self.uuid_to_key(table, '*', topic)
+            local_key = self._uuid_to_key(table, '*', topic)
             try:
                 ip_port = self.redis_mgt.get_ip_by_key(local_key)
                 client = self._get_client(local_key, ip_port)
@@ -269,12 +298,11 @@ class RedisDbDriver(db_api.DbApi):
         m = re.match(regex, key)
         return m.group(1)
 
-    def _allocate_unique_key(self):
-        local_key = self.uuid_to_key('tunnel_key', 'key', None)
+    def _allocate_unique_key(self, table):
+        local_key = self._uuid_to_key('unique_key', table, None)
         ip_port = None
         try:
-            ip_port = self.redis_mgt.get_ip_by_key(local_key)
-            client = self._get_client(local_key, ip_port)
+            client = self._update_client(local_key)
             if client is None:
                 return None
             return client.incr(local_key)
@@ -283,9 +311,9 @@ class RedisDbDriver(db_api.DbApi):
             LOG.exception(_LE("exception when incr: %(key)s, %(e)s")
                           % {'key': local_key, 'e': e})
 
-    def allocate_unique_key(self):
+    def allocate_unique_key(self, table):
         try:
-            return self._allocate_unique_key()
+            return self._allocate_unique_key(table)
         except Exception as e:
             LOG.error(_LE("allocate_unique_key exception: %(e)s")
                       % {'e': e})
@@ -294,23 +322,13 @@ class RedisDbDriver(db_api.DbApi):
     def register_notification_callback(self, callback, topics=None):
         pass
 
-    def uuid_to_key(self, table, key, topic):
-        if topic is None:
+    def _uuid_to_key(self, table, key, topic):
+        if not topic:
             local_key = ('{' + table + '.' + '}' + '.' + key)
         else:
             local_key = ('{' + table + '.' + topic + '}' + '.' + key)
         return local_key
 
-    def check_connection(self, ip_port):
-        try:
-            if self.clients[ip_port] is None:
-                return False
-            self.clients[ip_port].get(None)
-        except (exceptions.ConnectionError,
-                exceptions.BusyLoadingError):
-            return False
-        return True
-
     def _get_client(self, key=None, host=None):
         if host is None:
             ip_port = self.redis_mgt.get_ip_by_key(key)
@@ -326,10 +344,16 @@ class RedisDbDriver(db_api.DbApi):
             return None
 
     def process_ha(self):
-        self._update_server_list()
+        if self.is_neutron_server:
+            self._sync_master_list()
+        else:
+            self._update_server_list()
 
     def register_topic_for_notification(self, topic):
         pass
 
     def unregister_topic_for_notification(self, topic):
         pass
+
+    def set_neutron_server(self, is_neutron_server):
+        self.is_neutron_server = is_neutron_server
