@@ -24,10 +24,3 @@ class DFLockedObjects(model_base.BASEV2):
     lock = sa.Column(sa.Boolean, default=False)
     session_id = sa.Column(sa.BigInteger, default=0)
     created_at = sa.Column(sa.DateTime, onupdate=sa.func.utc_timestamp())
-
-
-class DFVersionObjects(model_base.BASEV2):
-    __tablename__ = 'dfversionobjects'
-
-    object_uuid = sa.Column(sa.String(36), primary_key=True)
-    version = sa.Column(sa.BigInteger, default=0)
