@@ -12,28 +12,31 @@
 
 import collections
 import fcntl
+import netaddr
 import os
 import re
 import socket
 import threading
 import time
 
-import eventlet
 from neutron.agent.common import utils
-from oslo_config import cfg
 from oslo_log import log
 import pytun
-import ryu.lib.packet
+from ryu.lib.packet import arp
+from ryu.lib.packet import dhcp
+from ryu.lib.packet import icmp
+from ryu.lib.packet import icmpv6
+from ryu.lib.packet import ipv4
+from ryu.lib.packet import ipv6
+from ryu.lib.packet import packet
 
 from dragonflow._i18n import _LI, _LE
-from dragonflow.common import common_params
 from dragonflow.common import utils as d_utils
+from dragonflow import conf as cfg
 from dragonflow.tests.common import utils as test_utils
 from dragonflow.tests.fullstack import test_objects as objects
 
 
-cfg.CONF.register_opts(common_params.df_opts, 'df')
-
 LOG = log.getLogger(__name__)
 
 
@@ -51,10 +54,14 @@ def create_tap_dev(dev, mac_address=None):
         # First, try with 'ip'
         utils.execute(['ip', 'tuntap', 'add', dev, 'mode', 'tap'],
                       run_as_root=True, check_exit_code=[0, 2, 254])
-    except Exception as e:
-        print e
-        # Second option: tunctl
-        utils.execute(['tunctl', '-b', '-t', dev], run_as_root=True)
+    except Exception:
+        try:
+            # Second option: tunctl
+            utils.execute(['tunctl', '-b', '-t', dev], run_as_root=True)
+        except Exception:
+            LOG.exception('Error while creating tap device {0}'.format(dev))
+            raise
+
     if mac_address:
         utils.execute(['ip', 'link', 'set', dev, 'address', mac_address],
                       run_as_root=True, check_exit_code=[0, 2, 254])
@@ -62,6 +69,24 @@ def create_tap_dev(dev, mac_address=None):
                   check_exit_code=[0, 2, 254])
 
 
+def delete_tap_device(dev):
+    """Delete a tap with name dev on the operating system.
+    :param dev: The name of the tap device to delete
+    :type dev:  String
+    """
+    try:
+        # First, try with 'ip'
+        utils.execute(['ip', 'tuntap', 'del', 'dev', dev, 'mode', 'tap'],
+                      run_as_root=True, check_exit_code=[0, 2, 254])
+    except Exception:
+        try:
+            # Second option: tunctl
+            utils.execute(['tunctl', '-d', dev], run_as_root=True)
+        except Exception:
+            LOG.exception('Error while deleting tap device {0}'.format(dev))
+            raise
+
+
 def packet_raw_data_to_hex(buf):
     return str(buf).encode('hex')
 
@@ -76,6 +101,8 @@ class Topology(object):
         self.neutron = neutron
         self.nb_api = nb_api
         self.network = objects.NetworkTestObj(neutron, nb_api)
+        self.external_network = objects.ExternalNetworkTestObj(neutron, nb_api)
+        self.exist_external_net = False
         self.subnets = []
         self.routers = []
         self.network.create()
@@ -108,6 +135,8 @@ class Topology(object):
             subnet.delete()
         self.subnets = []
         self.network.close()
+        if not self.exist_external_net:
+            self.external_network.close()
         self.fake_default_security_group.close()
 
     def close(self):
@@ -115,14 +144,18 @@ class Topology(object):
             self._is_closed = True
             self.delete()
 
-    def create_subnet(self, cidr='192.168.0.0/24'):
+    def create_subnet(self, cidr=None, enable_dhcp=True):
         """Create a subnet in this topology, with the given subnet address
         range.
-        :param cidr: The subnet's address range, in form <IP>/<mask len>
+        :param cidr: The subnet's address range, in form <IP>/<mask len>.
+                     If it is None, the cidr will be allocated from default
+                     subnetpool.
         :type cidr:  String
+        :param enable_dhcp: Whether to enable dhcp for this subnet.
+        :type cidr:  Boolean
         """
         subnet_id = len(self.subnets)
-        subnet = Subnet(self, subnet_id, cidr)
+        subnet = Subnet(self, subnet_id, cidr, enable_dhcp)
         self.subnets.append(subnet)
         return subnet
 
@@ -136,18 +169,40 @@ class Topology(object):
         self.routers.append(router)
         return router
 
+    def create_external_network(self, router_ids):
+        """Create external network in this topology, and use it as external
+        gateway to given routers.
+        """
+        external_net = objects.find_first_network(self.neutron,
+                                                  {'router:external': True})
+        if external_net:
+            self.exist_external_net = True
+            external_net_id = external_net['id']
+        else:
+            external_net_id = self.external_network.create()
+
+        for r in router_ids:
+            router = self.routers[r]
+            router.router.set_gateway(external_net_id)
+
+        return external_net_id
+
 
 class Subnet(object):
     """Represent a single subnet."""
-    def __init__(self, topology, subnet_id, cidr):
+    def __init__(self, topology, subnet_id, cidr, enable_dhcp):
         """Create the subnet under the given topology, with the given ID, and
         the given address range.
         :param topology:  The topology to which the subnet belongs
         :type topology:   Topology
         :param subnet_id: The subnet's ID in the topology. Created by topology
         :type subnet_id:  Number (Opaque)
-        :param cidr:      The address range for this subnet. Format IP/MaskLen
+        :param cidr:      The address range for this subnet. Format IP/MaskLen.
+                          If it is None, the cidr will be allocated from
+                          default subnetpool.
         :type cidr:       String
+        :param enable_dhcp: Whether to enable dhcp for this subnet.
+        :type cidr:  Boolean
         """
         self.topology = topology
         self.subnet_id = subnet_id
@@ -157,21 +212,19 @@ class Subnet(object):
             self.topology.nb_api,
             self.topology.network.network_id
         )
-        self.subnet.create(subnet={
-            'cidr': cidr,
-            'ip_version': 4,
-            'network_id': topology.network.network_id,
-            'host_routes': [
-                {
-                    'destination': '1.1.1.0/24',
-                    'nexthop': '2.2.2.2'
-                },
-                {
-                    'destination': '1.1.2.0/24',
-                    'nexthop': '3.3.3.3'
-                },
-            ]
-        })
+        if cidr:
+            ip_version = self._get_ip_version(cidr)
+            self.subnet.create(subnet={
+                'cidr': cidr,
+                'enable_dhcp': enable_dhcp,
+                'ip_version': ip_version,
+                'network_id': topology.network.network_id
+            })
+        else:
+            self.subnet.create()
+
+    def update(self, updated_parameters):
+        self.subnet.update(updated_parameters)
 
     def delete(self):
         """Delete this subnet, and all attached ports."""
@@ -196,6 +249,16 @@ class Subnet(object):
         self.ports.append(port)
         return port
 
+    def _get_ip_version(self, cidr):
+        """
+        Calculates the IP version from the CIDR, and returns it.
+        Raises AddrFormatError if the CIDR is not correctly formatted
+        :param cidr: The address range for this subnet. Format IP/MaskLen
+        """
+        network = netaddr.IPNetwork(cidr)
+        ip_version = network.ip.version
+        return ip_version
+
 
 class Port(object):
     """Represent a single port. Also contains access to the underlying tap
@@ -298,6 +361,7 @@ class LogicalPortTap(object):
             self.tap.fileno(),
         ))
         self.tap.close()
+        delete_tap_device(self.tap.name)
 
     def send(self, buf):
         """Send a packet out via the tap device.
@@ -378,6 +442,12 @@ class Router(object):
         self.router.close()
 
 
+class TimeoutException(Exception):
+
+    def __init__(self):
+        super(TimeoutException, self).__init__('Timeout')
+
+
 class Policy(object):
     """Represent a policy, i.e. the expected packets on each port in the
     topology, and the actions to take in each case.
@@ -449,7 +519,7 @@ class Policy(object):
         :param timeout: After this many seconds, throw an exception
         :type timeout:  Number
         """
-        exception = Exception('Timeout')
+        exception = TimeoutException()
         if timeout is not None:
             entry_time = time.time()
         for thread in self.threads:
@@ -565,52 +635,94 @@ class Filter(object):
 class RyuIPv6Filter(object):
     """Use ryu to parse the packet and test if it's IPv6."""
     def __call__(self, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
-        return (pkt.get_protocol(ryu.lib.packet.ipv6.ipv6) is not None)
+        pkt = packet.Packet(buf)
+        return (pkt.get_protocol(ipv6.ipv6) is not None)
+
+
+class RyuFilterIcmpv6ProtocolType(object):
+    """
+    Use ryu to parse the object and see if it from the requested icmpv6 type
+    """
+    type_ = icmpv6.icmpv6
+
+    def __call__(self, buf):
+        pkt = packet.Packet(buf)
+        pkt_protocol = pkt.get_protocol(icmpv6.icmpv6)
+        if not pkt_protocol:
+            return False
+        return pkt_protocol.type_ == self.type_
+
+
+class RyuNeighborAdvertisementFilter(RyuFilterIcmpv6ProtocolType):
+    """Use ryu to parse the packet and test if it's a Neighbor Advertisement"""
+    type_ = icmpv6.ND_NEIGHBOR_ADVERT
+
+
+class RyuIpv6MulticastFilter(RyuFilterIcmpv6ProtocolType):
+    """Use ryu to parse the object and see if it is a multicast request"""
+    type_ = icmpv6.MLDV2_LISTENER_REPORT
+
+
+class RyuARPRequestFilter(object):
+    """Use ryu to parse the packet and test if it's an ARP request."""
+    def __init__(self, arp_tpa=None):
+        self.arp_tpa = arp_tpa
+
+    def __call__(self, buf):
+        pkt = packet.Packet(buf)
+        pkt_arp_protocol = pkt.get_protocol(arp.arp)
+        if (not pkt_arp_protocol) or (
+                    pkt_arp_protocol.opcode != arp.ARP_REQUEST):
+            return False
+        if self.arp_tpa is not None:
+            return pkt_arp_protocol.dst_ip == self.arp_tpa
+
+        return True
 
 
 class RyuARPReplyFilter(object):
     """Use ryu to parse the packet and test if it's an ARP reply."""
     def __call__(self, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
-        arp = pkt.get_protocol(ryu.lib.packet.arp.arp)
-        if not arp:
+        pkt = packet.Packet(buf)
+        pkt_arp_protocol = pkt.get_protocol(arp.arp)
+        if not pkt_arp_protocol:
             return False
-        return arp.opcode == 2
+        return pkt_arp_protocol.opcode == 2
 
 
 class RyuARPGratuitousFilter(object):
     """Use ryu to parse the packet and test if it's a gratuitous ARP."""
     def __call__(self, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
-        arp = pkt.get_protocol(ryu.lib.packet.arp.arp)
-        if not arp:
+        pkt = packet.Packet(buf)
+        pkt_arp_protocol = pkt.get_protocol(arp.arp)
+        if not pkt_arp_protocol:
             return False
-        return arp.src_ip == arp.dst_ip
+        return pkt_arp_protocol.src_ip == pkt_arp_protocol.dst_ip
 
 
 # Taken from the DHCP app
 def _get_dhcp_message_type_opt(dhcp_packet):
     for opt in dhcp_packet.options.option_list:
-        if opt.tag == ryu.lib.packet.dhcp.DHCP_MESSAGE_TYPE_OPT:
+        if opt.tag == dhcp.DHCP_MESSAGE_TYPE_OPT:
             return ord(opt.value)
 
 
 class RyuDHCPFilter(object):
     """Use ryu to parse the packet and test if it's a DHCP Ack"""
     def __call__(self, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
-        return (pkt.get_protocol(ryu.lib.packet.dhcp.dhcp) is not None)
+        pkt = packet.Packet(buf)
+        return (pkt.get_protocol(dhcp.dhcp) is not None)
 
 
 class RyuDHCPPacketTypeFilter(object):
     """Use ryu to parse the packet and test if it's a DHCP Ack"""
     def __call__(self, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
-        dhcp = pkt.get_protocol(ryu.lib.packet.dhcp.dhcp)
-        if not dhcp:
+        pkt = packet.Packet(buf)
+        pkt_dhcp_protocol = pkt.get_protocol(dhcp.dhcp)
+        if not pkt_dhcp_protocol:
             return False
-        return _get_dhcp_message_type_opt(dhcp) == self.get_dhcp_packet_type()
+        dhcp_type = _get_dhcp_message_type_opt(pkt_dhcp_protocol)
+        return dhcp_type == self.get_dhcp_packet_type()
 
     def get_dhcp_packet_type(self):
         raise Exception('DHCP packet type filter not fully implemented')
@@ -618,23 +730,23 @@ class RyuDHCPPacketTypeFilter(object):
 
 class RyuDHCPOfferFilter(RyuDHCPPacketTypeFilter):
     def get_dhcp_packet_type(self):
-        return ryu.lib.packet.dhcp.DHCP_OFFER
+        return dhcp.DHCP_OFFER
 
 
 class RyuDHCPAckFilter(RyuDHCPPacketTypeFilter):
     def get_dhcp_packet_type(self):
-        return ryu.lib.packet.dhcp.DHCP_ACK
+        return dhcp.DHCP_ACK
 
 
 class RyuICMPFilter(object):
     def __call__(self, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
-        icmp = pkt.get_protocol(ryu.lib.packet.icmp.icmp)
-        if not icmp:
+        pkt = packet.Packet(buf)
+        pkt_icmp_protocol = pkt.get_protocol(icmp.icmp)
+        if not pkt_icmp_protocol:
             return False
-        return self.filter_icmp(pkt, icmp)
+        return self.filter_icmp(pkt, pkt_icmp_protocol)
 
-    def filter_icmp(self, pkt, icmp):
+    def filter_icmp(self, pkt, icmp_prot):
         return True
 
     def is_same_icmp(self, icmp1, icmp2):
@@ -657,13 +769,13 @@ class RyuICMPPingFilter(RyuICMPFilter):
         super(RyuICMPPingFilter, self).__init__()
         self.get_ping = get_ping
 
-    def filter_icmp(self, pkt, icmp):
-        if icmp.type != ryu.lib.packet.icmp.ICMP_ECHO_REQUEST:
+    def filter_icmp(self, pkt, proto):
+        if proto.type != icmp.ICMP_ECHO_REQUEST:
             return False
         result = True
         if self.get_ping is not None:
             ping = self.get_ping()
-            result = super(RyuICMPPingFilter, self).is_same_icmp(icmp, ping)
+            result = super(RyuICMPPingFilter, self).is_same_icmp(proto, ping)
         return result
 
 
@@ -677,11 +789,57 @@ class RyuICMPPongFilter(RyuICMPFilter):
         super(RyuICMPPongFilter, self).__init__()
         self.get_ping = get_ping
 
-    def filter_icmp(self, pkt, icmp):
-        if icmp.type != ryu.lib.packet.icmp.ICMP_ECHO_REPLY:
+    def filter_icmp(self, pkt, icmp_prot):
+        if icmp_prot.type != icmp.ICMP_ECHO_REPLY:
             return False
         ping = self.get_ping()
-        return super(RyuICMPPongFilter, self).is_same_icmp(icmp, ping)
+        return super(RyuICMPPongFilter, self).is_same_icmp(icmp_prot, ping)
+
+
+class RyuICMPTimeExceedFilter(RyuICMPFilter):
+    """
+    A filter to detect ICMP time exceed messages.
+    :param get_ip:    Return an object contained the original IP header
+    :type get_ip:     Callable with no arguments.
+    """
+    def __init__(self, get_ip):
+        super(RyuICMPTimeExceedFilter, self).__init__()
+        self.get_ip = get_ip
+
+    def filter_icmp(self, pkt, icmp_prot):
+        if icmp_prot.type != icmp.ICMP_TIME_EXCEEDED:
+            return False
+        ip_pkt = self.get_ip()
+        embedded_ip_pkt, c, p = ipv4.ipv4.parser(icmp_prot.data.data)
+        if ip_pkt.src != embedded_ip_pkt.src:
+            return False
+        if ip_pkt.dst != embedded_ip_pkt.dst:
+            return False
+
+        return True
+
+
+class RyuICMPUnreachFilter(RyuICMPFilter):
+    """
+    A filter to detect ICMP unreachable messages.
+    :param get_ip:    Return an object contained the original IP header
+    :type get_ip:     Callable with no arguments.
+    """
+    def __init__(self, get_ip):
+        super(RyuICMPUnreachFilter, self).__init__()
+        self.get_ip = get_ip
+
+    def filter_icmp(self, pkt, icmp_prot):
+        if icmp_prot.type != icmp.ICMP_DEST_UNREACH:
+            return False
+        ip_pkt = self.get_ip()
+        embedded_ip_pkt, c, p = ipv4.ipv4.parser(icmp_prot.data.data)
+        if ip_pkt.src != embedded_ip_pkt.src:
+            return False
+        if ip_pkt.dst != embedded_ip_pkt.dst:
+            return False
+
+        return True
 
 
 class Action(object):
@@ -705,7 +863,7 @@ class Action(object):
 class LogAction(Action):
     """Action to log the received packet."""
     def __call__(self, policy, rule, port_thread, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
+        pkt = packet.Packet(buf)
         LOG.info(_LI('LogAction: Got packet: {}').format(str(pkt)))
 
 
@@ -792,7 +950,7 @@ class RaiseAction(Action):
         self.message = message
 
     def __call__(self, policy, rule, port_thread, buf):
-        pkt = ryu.lib.packet.packet.Packet(buf)
+        pkt = packet.Packet(buf)
         raise Exception("Packet {} raised exception on port: {}: {}".format(
             str(pkt),
             (port_thread.port.subnet.subnet_id, port_thread.port.port_id),
@@ -832,7 +990,7 @@ class WaitAction(Action):
         self.wait_time = wait_time
 
     def __call__(self, policy, rule, port_thread, buf):
-        eventlet.sleep(self.wait_time)
+        time.sleep(self.wait_time)
 
 
 class PortThread(object):
