@@ -65,14 +65,14 @@ class TestNeutronAPIandDB(test_base.DFTestBase):
                 if port.get_device_owner() == 'network:dhcp':
                     dhcp_ports_found += 1
         network.close()
-        self.assertEqual(dhcp_ports_found, 1)
+        self.assertEqual(1, dhcp_ports_found)
         ports = self.nb_api.get_all_logical_ports()
         dhcp_ports_found = 0
         for port in ports:
             if port.get_lswitch_id() == network_id:
                 if port.get_device_owner() == 'network:dhcp':
                     dhcp_ports_found += 1
-        self.assertEqual(dhcp_ports_found, 0)
+        self.assertEqual(0, dhcp_ports_found)
 
     def test_create_delete_subnet(self):
         network = self.store(objects.NetworkTestObj(self.neutron, self.nb_api))
@@ -174,6 +174,69 @@ class TestNeutronAPIandDB(test_base.DFTestBase):
         network.close()
         self.assertFalse(network.exists())
 
+    def test_create_port_with_qospolicy(self):
+        network = self.store(objects.NetworkTestObj(self.neutron, self.nb_api))
+        network_id = network.create()
+        self.assertTrue(network.exists())
+
+        qospolicy = self.store(objects.QosPolicyTestObj(self.neutron,
+                                                        self.nb_api))
+        qos_policy_id = qospolicy.create()
+        self.assertTrue(qospolicy.exists())
+
+        port = self.store(objects.PortTestObj(self.neutron,
+                                              self.nb_api,
+                                              network_id))
+        port_param = {
+            'admin_state_up': True,
+            'name': 'port1',
+            'network_id': network_id,
+            'qos_policy_id': qos_policy_id
+        }
+        port.create(port_param)
+        self.assertTrue(port.exists())
+        self.assertEqual(qos_policy_id,
+                         port.get_logical_port().get_qos_policy_id())
+
+        port.close()
+        self.assertFalse(port.exists())
+        network.close()
+        self.assertFalse(network.exists())
+        qospolicy.close()
+        self.assertFalse(qospolicy.exists())
+
+    def test_update_port_with_qospolicy(self):
+        network = self.store(objects.NetworkTestObj(self.neutron, self.nb_api))
+        network_id = network.create()
+        self.assertTrue(network.exists())
+
+        qospolicy = self.store(objects.QosPolicyTestObj(self.neutron,
+                                                        self.nb_api))
+        qos_policy_id = qospolicy.create()
+        self.assertTrue(qospolicy.exists())
+
+        port = self.store(objects.PortTestObj(self.neutron,
+                                              self.nb_api,
+                                              network_id))
+        port.create()
+        self.assertTrue(port.exists())
+
+        port_param = {
+            'admin_state_up': True,
+            'name': 'port1',
+            'qos_policy_id': qos_policy_id
+        }
+        port.update(port_param)
+        self.assertEqual(qos_policy_id,
+                         port.get_logical_port().get_qos_policy_id())
+
+        port.close()
+        self.assertFalse(port.exists())
+        network.close()
+        self.assertFalse(network.exists())
+        qospolicy.close()
+        self.assertFalse(qospolicy.exists())
+
     def test_delete_router_interface_port(self):
         router = self.store(objects.RouterTestObj(self.neutron, self.nb_api))
         network = self.store(objects.NetworkTestObj(self.neutron, self.nb_api))
@@ -241,6 +304,17 @@ class TestNeutronAPIandDB(test_base.DFTestBase):
         secgroup.close()
         self.assertFalse(secgroup.exists())
 
+    def test_create_delete_qos_policy(self):
+        qospolicy = self.store(
+            objects.QosPolicyTestObj(self.neutron, self.nb_api))
+        policy_id = qospolicy.create()
+        self.assertTrue(qospolicy.exists())
+        rule = {'max_kbps': '1000', 'max_burst_kbps': '100'}
+        qospolicy.create_rule(policy_id, rule, 'bandwidth_limit')
+        self.assertTrue(qospolicy.exists())
+        qospolicy.close()
+        self.assertFalse(qospolicy.exists())
+
     @contextlib.contextmanager
     def _prepare_ext_net(self):
         external_net = objects.find_first_network(self.neutron,
@@ -501,8 +575,8 @@ class TestNeutronAPIandDB(test_base.DFTestBase):
         })
         lport = port.get_logical_port()
         self.assertIsNotNone(lport)
-        real_pairs = lport.get_allow_address_pairs()
-        self.assertEqual(expected_pairs, real_pairs)
+        real_pairs = lport.get_allowed_address_pairs()
+        self.assertItemsEqual(expected_pairs, real_pairs)
 
         expected_pairs = [
                 {"ip_address": "192.168.127.211",
@@ -513,5 +587,5 @@ class TestNeutronAPIandDB(test_base.DFTestBase):
         port.update({'allowed_address_pairs': expected_pairs})
         lport = port.get_logical_port()
         self.assertIsNotNone(lport)
-        real_pairs = lport.get_allow_address_pairs()
-        self.assertEqual(expected_pairs, real_pairs)
+        real_pairs = lport.get_allowed_address_pairs()
+        self.assertItemsEqual(expected_pairs, real_pairs)
