@@ -15,11 +15,20 @@
 
 ZUUL_CLONER=/usr/zuul-env/bin/zuul-cloner
 neutron_installed=$(echo "import neutron" | python 2>/dev/null ; echo $?)
+BRANCH_NAME=master
 
 set -ex
 
 cwd=$(/bin/pwd)
 
+CONSTRAINTS_FILE=$1
+shift
+
+install_cmd="pip install"
+if [ $CONSTRAINTS_FILE != "unconstrained" ]; then
+    install_cmd="$install_cmd -c$CONSTRAINTS_FILE"
+fi
+
 if [ $neutron_installed -eq 0 ]; then
     echo "ALREADY INSTALLED" > /tmp/tox_install.txt
     echo "Neutron already installed; using existing package"
@@ -29,16 +38,16 @@ elif [ -x "$ZUUL_CLONER" ]; then
     export ZUUL_BRANCH=${ZUUL_BRANCH-$BRANCH}
     $ZUUL_CLONER --cache-dir \
         /opt/git \
+        --branch $BRANCH_NAME \
         git://git.openstack.org \
         openstack/neutron
     cd openstack/neutron
-    pip install -e .
+    $install_cmd -e .
     cd "$cwd"
 else
     echo "PIP HARDCODE" > /tmp/tox_install.txt
-    pip install -U -egit+https://git.openstack.org/openstack/neutron#egg=neutron
+    $install_cmd -U -egit+https://git.openstack.org/openstack/neutron@$BRANCH_NAME#egg=neutron
 fi
 
-pip install -U $*
-pip install python-etcd
+$install_cmd -U $*
 exit $?
